import common_lib
import sys
import argparse
import datetime
import os
import redis
import random
import subprocess

MIN_RAND=1
MAX_RAND=10000

def main():
    args = parse_args()
    common_lib.init_rand(args.rand_seed)
    rc = redis.Redis(host=args.redishost, port=args.redisport, decode_responses=True)
    fun_name = args.proc_type + "_generator_env"
    common_lib.init_time(rc, False, fun_name)
    P = []
    active = 0
    overall = 0
    files = {}
    used = set()
    while True:
        work_done = False
        for i in range(len(P)):
            #poll() returns None if the process is still running, otherwise the exit code
            if P[i] != None and P[i].poll() != None:
                P[i] = None
                files[i].close()
                used.remove(i)
                print("DBG: Removed a process " + str([i, active, P]))
                sys.stdout.flush()
                active -= 1
        if args.proc_type != "customer":
            if args.min_wb < active <= args.max_wb and random.uniform(0, 1) > args.f_d:
                r = random.choice(list(used))
                if common_lib.SIMULATION_MODE:
                    print("DBG: Killing process " + str([common_lib.curr_elapsed_glob, common_lib.advance_elapsed_glob, r, active, args.min_wb, args.max_wb, overall, P]))
                else:
                    print("DBG: Killing process " + str([r, active, args.min_wb, args.max_wb, overall, P]))
                common_lib.redis_command(rc, "LPUSH " + args.queue_base + args.proc_type[0].upper() + str(r) + " TERM")
                work_done = True
        if active < args.min_wb or (active < args.max_wb and random.uniform(0, 1) > args.f_a):
            #cannot reuse indexes, one could find old non-processed items in its REDIS queue
            #furthermore, new channels is what actually happens...
            new_procs = max(1, args.min_wb - active)
            tbus = []
            for i in range(len(P)):
                if P[i] == None: #there was a proc, but it correctly ended
                    tbus += [i]
                    if len(tbus) == new_procs:
                        break
            for i in range(new_procs - len(tbus)): #for the remaining procs to be created
                P += [None]
                tbus += [len(P) - 1]
            for tbu in tbus:
                overall += 1
                # #necessary because, especially at the start, it is not aligned with other processes (it cannot be actually 0...)
                # common_lib.lump_time(rc, cause="before_creating") 
                common_lib.redis_command(rc, "LPUSH " + common_lib.QUEUE_INIT_BARR_SEND + " " + common_lib.QUEUE_PREFIX + args.proc_type + "_env_" + str(overall))
                active += 1
                files[tbu] = open(args.dir_for_log + "/" + args.proc_type + "_env_" + str(overall) + ".log", "w")
                used.add(tbu)
                work_done = True
                P[tbu] = subprocess.Popen(["/usr/bin/time", "-v", "-o", (args.dir_for_log + "/" + args.proc_type + "_env_" + str(overall) + ".time.log"), "python3", args.cmd_prefix + args.proc_type + "/src/main.py", "-i", str(overall), "-r", str(random.randint(MIN_RAND, MAX_RAND))] + args.other_args + [args.redishost, args.redisport], stdout=files[tbu], stderr=files[tbu])
                if common_lib.SIMULATION_MODE:
                    print("DBG: Generated new process " + str([common_lib.curr_elapsed_glob, common_lib.advance_elapsed_glob, tbu, active, args.min_wb, args.max_wb, overall, P]))
                else:
                    print("DBG: Generated new process " + str([tbu, active, args.min_wb, args.max_wb, overall, P]))
                sys.stdout.flush()
        if not(work_done):
            common_lib.sync_sleep(rc)
        common_lib.sleep_wrap(rc, cause="main_sleep")
        
def parse_args():
    QUEUE_BASE_DEF = "work:queue:T"
    PROC_TYPE_DEF = "customer"
    MIN_WB_DEF = 5
    MAX_WB_DEF = 10
    F_D_DEF = 0.5
    F_A_DEF = 2.0
    CMD_PREFIX_DEF = "environment/"
    OTHER_ARGS_DEF = ""
    RAND_SEED_DEF = 10
    DIR_FOR_LOG_DEF = "./"

    parser = argparse.ArgumentParser(description='Command arguments')
    parser.add_argument("-r", "--rand_seed", dest = "rand_seed", help = "rand_seed value (default is " + str(RAND_SEED_DEF) + ")", default = RAND_SEED_DEF, type = int)
    parser.add_argument("-d", "--dir_for_log", dest = "dir_for_log", help = "dir_for_log value (default is " + str(DIR_FOR_LOG_DEF) + ")", default = DIR_FOR_LOG_DEF, type = str)
    parser.add_argument("-q", "--queue_base", dest = "queue_base", help = "queue_base value (default is " + str(QUEUE_BASE_DEF) + ")", default = QUEUE_BASE_DEF, type = str)
    parser.add_argument("-o", "--other_args", dest = "other_args", help = "other_args value (default is " + str(OTHER_ARGS_DEF) + ")", default = OTHER_ARGS_DEF, type = str)
    parser.add_argument("-c", "--cmd_prefix", dest = "cmd_prefix", help = "cmd_prefix value (default is " + str(CMD_PREFIX_DEF) + ")", default = CMD_PREFIX_DEF, type = str)
    parser.add_argument("-p", "--proc_type", dest = "proc_type", help = "proc_type value (default is " + str(PROC_TYPE_DEF) + ")", default = PROC_TYPE_DEF, type = str)
    parser.add_argument("-w", "--min_wb", dest = "min_wb", help = "min_wb value (default is " + str(MIN_WB_DEF) + ")", default = MIN_WB_DEF, type = int)
    parser.add_argument("-W", "--max_wb", dest = "max_wb", help = "max_wb value (default is " + str(MAX_WB_DEF) + ")", default = MAX_WB_DEF, type = int)
    parser.add_argument("-f", "--f_d", dest = "f_d", help = "f_d value (default is " + str(F_D_DEF) + ")", default = F_D_DEF, type = float)
    parser.add_argument("-F", "--f_a", dest = "f_a", help = "f_a value (default is " + str(F_A_DEF) + ")", default = F_A_DEF, type = float)
    parser.add_argument("redishost", help = "redishost", type = str)
    parser.add_argument("redisport", help = "redisport", type = str)

    common_lib.print_cmd_line(sys.argv)
    args = parser.parse_args()

    if args.dir_for_log == "":
        print("Error: dir_for_log (-d) must not be the empty string")
        sys.exit(2)

    if args.queue_base == "":
        print("Error: queue_base (-q) must not be the empty string")
        sys.exit(2)

    if args.cmd_prefix == "":
        print("Error: cmd_prefix (-c) must not be the empty string")
        sys.exit(2)

    if args.proc_type not in ["customer", "producer", "shipper"]:
        print("Error: proc_type (-p) must be wither customer, producer, or shipper")
        sys.exit(2)

    if args.min_wb <= 0:
        print("Error: min_wb (-w) must be positive")
        sys.exit(2)

    if args.f_a <= 0:
        print("Error: f_a (-f) must be positive")
        sys.exit(2)

    if args.f_d <= 0:
        print("Error: f_d (-F) must be positive")
        sys.exit(2)

    if args.max_wb < args.min_wb:
        print("Error: max_wb (-W) must not be less than min_wb (-w)")
        sys.exit(2)

    #needed DELETE_THIS as -opt would be interpreted as another option for this program...
    if args.other_args != "":
        args.other_args = args.other_args.replace("DELETE_THIS", "").split("|")
    else:
        args.other_args = []
        
    common_lib.print_all_options(args)
    return args
    
main()
