#include <hiredis/hiredis.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "aux.h"
#include <assert.h>
#include <math.h>
#include "clib.h"
#include <pqxx/pqxx>

#define DB_ONLINE_STATS 0
#define BUF_SIZE (1 << 21)
static char sql[BUF_SIZE];

static void report(redisContext *conn, pqxx::connection *C, const char *reason, FILE *f)
{
  int *save, i = 0;
  {
    pqxx::nontransaction N(*C);
    pqxx::result *r = db_cmd(conn, (pqxx::work *)0, &N, 0 + 4*DB_ONLINE_STATS, sql);
    save = (int *)calloc(r->size(), sizeof(int));
    for (pqxx::result::const_iterator c = r->begin(); c != r->end(); ++c) { 
      fprintf(f, "%d|%d|%d|%d|%s|%s|%s|%s|%s|%s\n", c[0].as<int>(), c[1].as<int>(), c[2].as<int>(), c[3].as<int>(), c[4].as<std::string>().c_str(), c[5].is_null()? "NULL" : c[5].as<std::string>().c_str(), c[6].is_null()? "NULL" : c[6].as<std::string>().c_str(), c[7].is_null()? "NULL" : c[7].as<std::string>().c_str(), c[8].is_null()? "NULL" : c[8].as<std::string>().c_str(), reason);
      fflush(f);
      save[i++] = c[0].as<int>();
    }
    delete r;
  }
  for (int j = 0; j < i; j++){
    assert(sprintf(sql, "UPDATE order_from_customer SET reported = TRUE WHERE id = %d", save[j]) < BUF_SIZE);
    pqxx::work W(*C);
    db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
  }
}

void aux_monitor(redisContext *conn, pqxx::connection *C, FILE *f, double t)
{
  double elapsed = 0;
  assert(sprintf(sql, "SELECT id, customer_id, item_id, how_many, date_placed, date_processed, date_ready, date_shipped, date_arrived FROM order_from_customer WHERE date_arrived IS NOT NULL AND (date_arrived - date_placed > INTERVAL '%lf seconds' OR date_arrived - date_placed < INTERVAL '0 seconds') AND NOT(reported)", t) < BUF_SIZE);
  report(conn, C, "arr-placed>t", f);
  assert(sprintf(sql, "SELECT id, customer_id, item_id, how_many, date_placed, date_processed, date_ready, date_shipped, date_arrived FROM order_from_customer WHERE %s - date_placed > INTERVAL '%lf seconds' AND NOT(reported)", get_curr_time_psql(), t) < BUF_SIZE);
  report(conn, C, "now-placed>t", f);
}
