#include <hiredis/hiredis.h>
#include <string.h>
#include <stdlib.h>
#include "aux.h"
#include <assert.h>
#include <math.h>
#include "clib.h"
#include <pqxx/pqxx> 

#define DBG_HERE 0
#define DBG_HERE2 0
#define DB_ONLINE_STATS 0
#define BUF_SIZE (1 << 21)

static void send_prod(redisContext *conn, pqxx::connection *C, int *item_ids, int *ms, unsigned size_in, char *sql);
static pqxx::result *get_result(redisContext *conn, pqxx::connection *C, int type, char *sql);
static redisReply **fill_further_ids(redisContext *conn, pqxx::connection *C, int type, unsigned *further_ids_len);
static char sql[BUF_SIZE];
static char send[BUF_SIZE];

static void customer_req(redisContext *conn, pqxx::connection *C, char *sql, char *send)
{
  char *msg[5], *save_ptr[2];
  unsigned how_many = 0;
  for (msg[4] = strtok_r(send, "|", &save_ptr[0]); msg[4]; msg[4] = strtok_r((char *)0, "|", &save_ptr[0])) {
    msg[1] = strtok_r(msg[4], " ", &save_ptr[1]);
    msg[2] = strtok_r((char *)0, " ", &save_ptr[1]);
    if (!msg[2]) {
      how_many = atoi(msg[1]);
      assert(sprintf(sql, "INSERT INTO order_from_customer(customer_id, item_id, how_many, date_placed) VALUES ") < BUF_SIZE);
    }
    else {
      msg[3] = strtok_r((char *)0, " ", &save_ptr[1]);
      if (msg[3]) {
	char *send_back = strtok_r((char *)0, " ", &save_ptr[1]);
	assert(sprintf(&sql[strlen(sql)], "%s(%s, %s, %s, %s)", sql[strlen(sql) - 1] == ' '? "" : ", ", msg[3], msg[1], msg[2], get_curr_time_psql()) < BUF_SIZE - strlen(sql));
	// to be precise, the order id should be sent back, but we would need to modify the DB structure, I prefer to avoid this; in any case, customer script discards the value
	how_many--;
	if (how_many == 0) {
	  pqxx::work W(*C);
	  db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
	  redisReply *reply = (redisReply *)redis_cmd(conn, "LPUSH %s%s OK", QUEUE_CI_BASE, send_back);
	  freeReplyObject(reply);
	}
      }
      else {
	assert(sprintf(sql, "INSERT INTO customer (descr, arrived) VALUES ('%s', %s) RETURNING id", msg[2], get_curr_time_psql()) < BUF_SIZE);
	int cust_id;
	{
	  pqxx::work W(*C);
	  pqxx::result *r = db_cmd(conn, &W, (pqxx::nontransaction *)0, 3 + 4*DB_ONLINE_STATS, sql);
	  cust_id = (*r)[0][0].as<int>();
	  delete r;
	}
	assert(sprintf(sql, "SELECT id, how_many FROM item") < BUF_SIZE);
	pqxx::nontransaction N(*C);
	pqxx::result *r = db_cmd(conn, (pqxx::work *)0, &N, 0 + 4*DB_ONLINE_STATS, sql);
	assert(sprintf(sql, "%d ", cust_id) < BUF_SIZE);
	for (pqxx::result::const_iterator c = r->begin(); c != r->end(); ++c) 
	  assert(sprintf(&sql[strlen(sql)], "%d %d ", c[0].as<int>(), c[1].as<int>()) < BUF_SIZE - strlen(sql));
	delete r;
	redisReply *reply = (redisReply *)redis_cmd(conn, "LPUSH %s%s %s", QUEUE_CI_BASE, msg[2], sql);
	freeReplyObject(reply);
      }
    }
  }
}

static void producer_req(redisContext *conn, pqxx::connection *C, char *sql, char *send)
{
  char *tmp4, *save_ptr[2];
  for (tmp4 = strtok_r(send, "|", &save_ptr[0]); tmp4; tmp4 = strtok_r((char *)0, "|", &save_ptr[0])) {
    char *tmp2 = strtok_r(tmp4, " ", &save_ptr[1]);
    if (strcmp(tmp2, "init") == 0) {
      tmp2 = strtok_r((char *)0, " ", &save_ptr[1]);
      assert(sprintf(sql, "INSERT INTO producer (channel_code, arrived) VALUES(%s, %s)", tmp2, get_curr_time_psql()) < BUF_SIZE);
      pqxx::work W(*C);
      db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
      assert(sprintf(sql, "SELECT id FROM item") < BUF_SIZE);
      pqxx::nontransaction N(*C);
      pqxx::result *r = db_cmd(conn, (pqxx::work *)0, &N, 0 + 4*DB_ONLINE_STATS, sql);
      sql[0] = '\0';
      for (pqxx::result::const_iterator c = r->begin(); c != r->end(); ++c) 
	assert(sprintf(&sql[strlen(sql)], "%d ", c[0].as<int>()) < BUF_SIZE - strlen(sql));
      delete r;
      redisReply *reply = (redisReply *)redis_cmd(conn, "LPUSH %s%s %s", QUEUE_PI_BASE, tmp2, sql);
      freeReplyObject(reply);
    }
    else if (strcmp(tmp2, "term") == 0) {
      tmp2 = strtok_r((char *)0, " ", &save_ptr[1]);
      // not actually needed: channel_code's are always generated differently; stays are historical record
      // sprintf(sql, "DELETE FROM item_producer WHERE producer_id = %s", tmp2);
      pqxx::work W(*C);
      // db_cmd(conn, &W, (pqxx::nontransaction *)0, 0 + 4*DB_ONLINE_STATS, sql);
      assert(sprintf(sql, "UPDATE producer SET terminated = %s WHERE channel_code = %s AND terminated IS NULL", get_curr_time_psql(), tmp2) < BUF_SIZE);
      db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
    }
    else if (strcmp(tmp2, "prod") == 0) {
      char *prod_id = strtok_r((char *)0, " ", &save_ptr[1]);
      assert(sprintf(sql, "INSERT INTO item_producer (producer_id, item_id) VALUES ") < BUF_SIZE);
      int first = 1;
      for (tmp2 = strtok_r((char *)0, " ", &save_ptr[1]); tmp2; first = 0, tmp2 = strtok_r((char *)0, " ", &save_ptr[1])) 
	assert(sprintf(&sql[strlen(sql)], "%s(%s, %s)", first? "" : ", ", prod_id, tmp2) < BUF_SIZE - strlen(sql));
      pqxx::work W(*C);
      db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
    }
    else {
      char *how_many = strtok_r((char *)0, " ", &save_ptr[1]);
      assert(sprintf(sql, "UPDATE item SET (producing, how_many) = (false, how_many + %s) WHERE id = %s", how_many, tmp2) < BUF_SIZE);
      pqxx::work W(*C);
      db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
    }
  }
}

static void shipper_req(redisContext *conn, pqxx::connection *C, char *sql, char *send)
{
  char *tmp4, *save_ptr[2];
  for (tmp4 = strtok_r(send, "|", &save_ptr[0]); tmp4; tmp4 = strtok_r((char *)0, "|", &save_ptr[0])) {
    char *ord_id = strtok_r(tmp4, " ", &save_ptr[1]);
    if (strcmp(ord_id, "init") == 0) {
      char *tmp2 = strtok_r((char *)0, " ", &save_ptr[1]);
      assert(sprintf(sql, "INSERT INTO shipper (channel_code, arrived, curr_sending) VALUES(%s, %s, 0)", tmp2, get_curr_time_psql()) < BUF_SIZE);
      pqxx::work W(*C);
      db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
    }
    else if (strcmp(ord_id, "term") == 0) {
      char *tmp2 = strtok_r((char *)0, " ", &save_ptr[1]);
      assert(sprintf(sql, "UPDATE shipper SET terminated = %s WHERE channel_code = %s AND terminated IS NULL", get_curr_time_psql(), tmp2) < BUF_SIZE);
      pqxx::work W(*C);
      db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
    }
    else {
      char *tmp2 = strtok_r((char *)0, " ", &save_ptr[1]);
      assert(sprintf(sql, "UPDATE order_from_customer SET date_arrived = %s WHERE id = %s", get_curr_time_psql(), ord_id) < BUF_SIZE);
      pqxx::work W(*C);
      db_cmd(conn, &W, (pqxx::nontransaction *)0, 0 + 4*DB_ONLINE_STATS, sql);
      assert(sprintf(sql, "UPDATE shipper SET curr_sending = curr_sending - 1 WHERE channel_code = %s", tmp2) < BUF_SIZE);
      db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
    }
  }
}

static void man_ord_req(redisContext *conn, pqxx::connection *C, char *sql, pqxx::result *res_all)
{
  if (res_all->size() == 0) {
    delete res_all;
    return;
  }
  char *all_ord, *affected_ord, *for_strtok_r;
  unsigned i, num_items, *s, *t;
  {
    sql[0] = '\0';
    num_items = 0;
    i = 0;
    while (i < res_all->size()) {
      assert(sprintf(&sql[strlen(sql)], "%s%d", i == 0? "" : ",", (*res_all)[i][1].as<int>()) < BUF_SIZE - strlen(sql));
      if ((*res_all)[i][2].as<int>() <= (*res_all)[i][3].as<int>()) {
	num_items++;
	unsigned j;
	for (j = i + 1; j < res_all->size() && (*res_all)[j][0].as<int>() == (*res_all)[i][0].as<int>(); j++);
	i = j;
      }
      else
	i++;
    }
    all_ord = strdup(sql); // possibly with repeated ids, but no harm
    if (num_items > 0) {
      s = (unsigned *)calloc(num_items, sizeof(unsigned));
      t = (unsigned *)calloc(num_items, sizeof(unsigned));
      i = 0;
      unsigned for_st = 0;
      sql[0] = '\0';
      while (i < res_all->size()) {
	if ((*res_all)[i][2].as<int>() <= (*res_all)[i][3].as<int>()) {
	  s[for_st] = (*res_all)[i][3].as<int>() - (*res_all)[i][2].as<int>();
	  t[for_st] = (*res_all)[i][0].as<int>();
	  assert(sprintf(&sql[strlen(sql)], "%s%d", sql[0] == '\0'? "" : ",", (*res_all)[i][1].as<int>()) < BUF_SIZE - strlen(sql));
	  unsigned j;
	  for (j = i + 1; j < res_all->size() && (*res_all)[j][0].as<int>() == (*res_all)[i][0].as<int>(); j++) {
	    if ((*res_all)[j][2].as<int>() <= (*res_all)[j][3].as<int>() && s[for_st] >= (*res_all)[j][2].as<int>()) {
	      s[for_st] -= (*res_all)[j][2].as<int>();
	      assert(sprintf(&sql[strlen(sql)], ",%d", (*res_all)[j][1].as<int>()) < BUF_SIZE - strlen(sql));
	    }
	  }
	  i = j;
	  for_st++;
	}
	else
	  i++;
      }
      affected_ord = strdup(sql);
    }
    delete res_all;
  }
  {
    assert(sprintf(sql, "UPDATE order_from_customer SET date_processed = %s WHERE id IN (%s) AND date_processed IS NULL", get_curr_time_psql(), all_ord) < BUF_SIZE);
    pqxx::work W(*C);
    db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
  }
  if (num_items > 0) {
    {
      assert(sprintf(sql, "WITH tmp_tab(item_id, new_hm) AS (VALUES ") < BUF_SIZE);
      for (i = 0; i < num_items; i++) 
	assert(sprintf(&sql[strlen(sql)], "%s(%d,%d)", i == 0? "" : ",", t[i], s[i]) < BUF_SIZE - strlen(sql));
      free(s);
      free(t);
      assert(sprintf(&sql[strlen(sql)], ") UPDATE item SET how_many = new_hm FROM tmp_tab WHERE id = item_id") < BUF_SIZE - strlen(sql));
      pqxx::work W(*C);
      db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
    }
    {
      assert(sprintf(sql, "UPDATE order_from_customer SET date_ready = %s WHERE id IN (%s)", get_curr_time_psql(), affected_ord) < BUF_SIZE);
      pqxx::work W(*C);
      db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
    }
    {
      assert(sprintf(sql, "SELECT item_id, o.how_many, it.how_many FROM order_from_customer o INNER JOIN item it ON it.id = o.item_id WHERE o.id IN (%s) AND o.id NOT IN (%s) ORDER BY item_id", all_ord, affected_ord) < BUF_SIZE);
      free(affected_ord);
      pqxx::work W(*C);
      pqxx::result *r = db_cmd(conn, &W, (pqxx::nontransaction *)0, 3 + 4*DB_ONLINE_STATS, sql);
      if (r->size() > 0) {
	unsigned how_many_items = 1;
	int prev = (*r)[0][0].as<int>();
	for (i = 1; i < r->size(); i++) {
	  if ((*r)[i][0].as<int>() != prev)
	    how_many_items++;
	  prev = (*r)[i][0].as<int>();
	}
	int *tbp1, *tbp2;
	tbp1 = (int *)calloc(how_many_items, sizeof(int));
	tbp2 = (int *)calloc(how_many_items, sizeof(int));
	unsigned j = 0;
	tbp1[j] = (*r)[0][0].as<int>();
	tbp2[j] = (*r)[0][1].as<int>() - ((*r)[0][2].as<int>());
	for (i = 1; i < r->size(); i++) {
	  if ((*r)[i][0].as<int>() != tbp1[j]) {
	    tbp1[++j] = (*r)[i][0].as<int>();
	    tbp2[j] = (*r)[i][1].as<int>() - ((*r)[i][2].as<int>());
	  }
	  else
	    tbp2[j] += (*r)[i][1].as<int>();
	}
	assert(how_many_items == j + 1);
	send_prod(conn, C, tbp1, tbp2, how_many_items, sql);
	free(tbp1);
	free(tbp2);
      }
      delete r;
    }
  }
  free(all_ord);
}

static void send_ord_req(redisContext *conn, pqxx::connection *C, char *sql, pqxx::result *res_all, redisReply **further_ids, unsigned further_ids_len)
{
  if (res_all->size() == 0 && further_ids_len == 0) {
    delete res_all;
    return;
  }
  char *tmp2, *for_strtok_r;
  long long int size_all = res_all->size();
  unsigned i;
  int max_item_ind = 0;
  for (i = 0; i < further_ids_len; i++)
    size_all += further_ids[i]->elements;
  int **res = (int **)calloc(size_all, sizeof(int *));
  sql[0] = '\0';
  for (i = 0; i < res_all->size(); i++) {
    res[i] = (int *)calloc(3, sizeof(int));
    res[i][0] = (*res_all)[i][0].as<int>();
    res[i][1] = (*res_all)[i][1].as<int>();
    res[i][2] = (*res_all)[i][2].as<int>();
    if (max_item_ind < res[i][0])
      max_item_ind = res[i][0];
    assert(sprintf(&sql[strlen(sql)], "%s%d", sql[0] == '\0'? "" : ",", res[i][1]) < BUF_SIZE - strlen(sql));
  }
  char *tmp1 = strdup(sql);
  for (i = 0; i < further_ids_len; i++) {
    for (unsigned c = 0; c < further_ids[i]->elements; c++) {
      res[i] = (int *)calloc(3, sizeof(int));
      char *tmp[5];
      tmp[0] = strdup(further_ids[i]->element[c]->str);
      tmp[1] = strtok_r(tmp[0], " ", &tmp[4]);
      tmp[2] = strtok_r((char *)0, " ", &tmp[4]);
      tmp[3] = strtok_r((char *)0, " ", &tmp[4]);
      res[i + res_all->size()][0] = atoi(tmp[1]);
      res[i + res_all->size()][1] = atoi(tmp[2]);
      res[i + res_all->size()][2] = atoi(tmp[3]);
      if (max_item_ind < res[i + res_all->size()][0])
	max_item_ind = res[i + res_all->size()][0];
      free(tmp[0]);
    }
  }
  delete res_all;
  unsigned *S = (unsigned *)calloc(size_all, sizeof(unsigned));
  int max_ship_ind = 0;
  i = 0;
  while (i < size_all) {
    assert(sprintf(sql, "UPDATE shipper SET curr_sending = curr_sending + 1 WHERE id = (SELECT id FROM shipper WHERE curr_sending = (SELECT min(curr_sending) FROM shipper WHERE terminated IS NULL) AND terminated IS NULL LIMIT 1) RETURNING channel_code") < BUF_SIZE);
    pqxx::work W(*C);
    pqxx::result *r = db_cmd(conn, &W, (pqxx::nontransaction *)0, 3 + 4*DB_ONLINE_STATS, sql);
    // lump_time(conn, "update"); //inside db_cmd
    lump_time(conn, "select", (struct timespec *)0);
    lump_time(conn, "select", (struct timespec *)0);
    if (r->size() == 1) {
      pqxx::result::const_iterator c = r->begin();
      S[i] = c[0].as<int>();
      if (max_ship_ind < S[i])
	max_ship_ind = S[i];
      i++;
    }
    else
      // all shippers have terminated, thus also previous selected shippers are not valid any more...
      // it still may happen that some shipper terminate after having been selected; in that case, the next round of WB will detect it...
      i = 0;
    delete r;
  }
  struct pair { int o, n; } **to_be_sent;
  to_be_sent = (struct pair **)calloc(max_ship_ind, sizeof(struct pair *)); // indices start from 1...
  for (unsigned tbs = 0; tbs < max_ship_ind; tbs++)
    to_be_sent[tbs] = (struct pair *)calloc(max_item_ind, sizeof(struct pair));
  for (i = 0; i < size_all; i++) {
    to_be_sent[S[i] - 1][res[i][0] - 1].o = res[i][1];
    to_be_sent[S[i] - 1][res[i][0] - 1].n = res[i][2];
  }
  free(S);
  if (tmp1[0] != '\0') {
    assert(sprintf(sql, "UPDATE order_from_customer SET date_shipped = %s WHERE id IN (%s)", get_curr_time_psql(), tmp1) < BUF_SIZE);
    pqxx::work W(*C);
    db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
  }
  free(tmp1);
  for (unsigned tbs = 0; tbs < max_ship_ind; tbs++) {
    sql[0] = '\0';
    for (unsigned it = 0; it < max_item_ind; it++) {
      if (to_be_sent[tbs][it].n > 0)
	assert(sprintf(&sql[strlen(sql)], "%d %d %d ", it + 1, to_be_sent[tbs][it].o, to_be_sent[tbs][it].n) < BUF_SIZE - strlen(sql));
    }
    if (sql[0] != '\0') {
      redisReply *reply = (redisReply *)redis_cmd(conn, "LPUSH %s%d %s", QUEUE_SI_BASE, tbs + 1, sql);
      freeReplyObject(reply);
    }
  }
  for (unsigned tbs = 0; tbs < max_ship_ind; tbs++)
    free(to_be_sent[tbs]);
  free(to_be_sent);
  for (i = 0; i < size_all; i++)
    free(res[i]);
  free(res);
}

static void man_min_stor_req(redisContext *conn, pqxx::connection *C, char *sql, pqxx::result *res_all, redisReply **further_ids, unsigned further_ids_len)
{
  if (res_all->size() == 0 && further_ids_len == 0) {
    delete res_all;
    return;
  }
  char *tmp2, *save[2];
  long long int size_all = res_all->size();
  unsigned i;
  for (i = 0; i < further_ids_len; i++)
    size_all += further_ids[i]->elements;
  int *its = (int *)calloc(size_all, sizeof(int));
  int *nums = (int *)calloc(size_all, sizeof(int));
  for (i = 0; i < res_all->size(); i++) {
    its[i] = (*res_all)[i][0].as<int>();
    nums[i] = (*res_all)[i][1].as<int>();
  }
  for (i = 0; i < further_ids_len; i++) {
    for (unsigned c = 0; c < further_ids[i]->elements; c++) {
      char *tmp[4];
      tmp[0] = strdup(further_ids[i]->element[c]->str);
      tmp[1] = strtok_r(tmp[0], " ", &tmp[3]);
      tmp[2] = strtok_r((char *)0, " ", &tmp[3]);
      its[i + res_all->size()] = atoi(tmp[1]);
      nums[i + res_all->size()] = atoi(tmp[2]);
      free(tmp[0]);
    }
  }
  delete res_all;
  send_prod(conn, C, its, nums, size_all, sql);
  free(its);
  free(nums);
}

void manage_reqs(redisContext *conn, pqxx::connection *C, int type)
{
  const char *ch;
  pqxx::result *res;
  unsigned further_ids_len;
  redisReply **further_ids;
  if (type == 0)
    ch = QUEUE_CO;
  else if (type == 1)
    ch = QUEUE_PO;
  else if (type == 2)
    ch = QUEUE_SO;
  if (type < 3) {
    redisReply *reply = (redisReply *)redis_cmd(conn, "BRPOP %s 0", ch);
    assert(sprintf(send, "%s|", reply->element[1]->str) < BUF_SIZE);
    freeReplyObject(reply);
    reply = (redisReply *)redis_cmd(conn, "LLEN %s", ch);
    long long int read_num = reply->integer;
    freeReplyObject(reply);
    if (read_num > 0) {
      reply = (redisReply *)redis_cmd(conn, "RPOP %s %ld", ch, read_num);
      for (unsigned i = 0; i < read_num; i++) {
	assert(strlen(send) + strlen(reply->element[i]->str) + 1 + 1 < BUF_SIZE);
	strcat(send, reply->element[i]->str);
	strcat(send, "|");
      }
      freeReplyObject(reply);
    }
  }
  else {
    res = get_result(conn, C, type, sql);
    if (type == 4 || type == 5) {
      further_ids = fill_further_ids(conn, C, type, &further_ids_len);
      if (further_ids_len > 0) {
	printf("%d futher ids\n", further_ids_len);
	fflush(stdout);
      }
    }
  }
  if (type == 0)
    customer_req(conn, C, sql, send);
  else if (type == 1)
    producer_req(conn, C, sql, send);
  else if (type == 2)
    shipper_req(conn, C, sql, send);
  else if (type == 3)
    man_ord_req(conn, C, sql, res);
  else if (type == 4)
    man_min_stor_req(conn, C, sql, res, further_ids, further_ids_len);
  else
    send_ord_req(conn, C, sql, res, further_ids, further_ids_len);
  if (type >= 4) {
    // delete res; //done inside functions
    for (unsigned f = 0; f < further_ids_len; f++)
      freeReplyObject(further_ids[f]);
    free(further_ids);
  }
}

static void send_prod(redisContext *conn, pqxx::connection *C, int *item_ids, int *ms, unsigned size_in, char *sql)
{
  char *items = (char *)0;
  int max_item_ind = item_ids[0];
  {
    assert(sprintf(sql, "UPDATE item SET producing = true WHERE NOT(producing) AND id IN (%d", item_ids[0]) < BUF_SIZE);
    for (unsigned i = 1; i < size_in; i++) {
      if (max_item_ind < item_ids[i])
	max_item_ind = item_ids[i];
      assert(sprintf(&sql[strlen(sql)], ",%d", item_ids[i]) < BUF_SIZE - strlen(sql));
    }
    assert(sprintf(&sql[strlen(sql)], ") RETURNING id") < BUF_SIZE - strlen(sql));
    pqxx::work W(*C);
    pqxx::result *r = db_cmd(conn, &W, (pqxx::nontransaction *)0, 3 + 4*DB_ONLINE_STATS, sql);
    if (r->size() > 0) {
      sql[0] = '\0';
      for (unsigned i = 0; i < r->size(); i++)
	assert(sprintf(&sql[strlen(sql)], "%s%d", i == 0? "" : ",", (*r)[i][0].as<int>()) < BUF_SIZE - strlen(sql));
      items = strdup(sql);
      delete r;
    }
    else {
      delete r;
      return;
    }
  }
  assert(sprintf(sql, "SELECT item_id, producer_id FROM item_producer INNER JOIN producer ON producer_id = channel_code WHERE item_id IN (%s) AND terminated IS NULL ORDER BY item_id", items) < BUF_SIZE);
  free(items);
  pqxx::nontransaction N(*C);
  pqxx::result *r = db_cmd(conn, (pqxx::work *)0, &N, 0 + 4*DB_ONLINE_STATS, sql);
  if (r->size() == 0) { //very unlikely, No producers for any item!
    delete r;
    return;
  }
  int max_prod_ind = (*r)[0][1].as<int>(), row;
  for (row = 1; row < r->size(); row++) {
    if (max_prod_ind < (*r)[row][1].as<int>())
      max_prod_ind = (*r)[row][1].as<int>();
  }
  unsigned **to_be_sent = (unsigned **)calloc(max_prod_ind, sizeof(unsigned *)); // indices start from 1...
  for (unsigned tbs = 0; tbs < max_prod_ind; tbs++)
    to_be_sent[tbs] = (unsigned *)calloc(max_item_ind, sizeof(unsigned));
  for (unsigned it = 0; it < size_in; it++) {
    unsigned producer_ids_len = 0;
    for (row = 0; row < r->size(); row++) {
      if ((*r)[row][0].as<int>() == item_ids[it])
	producer_ids_len++;
      else if (producer_ids_len > 0)
	break;
    }
    unsigned *producer_ids = (unsigned *)calloc(producer_ids_len, sizeof(unsigned)), idx = 0;
    for (row = 0; row < r->size(); row++) {
      if ((*r)[row][0].as<int>() == item_ids[it])
	producer_ids[idx++] = (*r)[row][1].as<int>();
      else if (idx > 0)
	break;
    }
    if (producer_ids_len == 0) {
      free(producer_ids);
      continue;
    }
    unsigned i, m = ms[it];
    unsigned a = m/producer_ids_len;
    unsigned b = m%producer_ids_len;
    if (a != 0) {
      for (i = 0; i < producer_ids_len; i++)
	to_be_sent[producer_ids[i] - 1][item_ids[it] - 1] += a;
    }
    for (i = 0; i < b; i++) 
      to_be_sent[producer_ids[i] - 1][item_ids[it] - 1] ++;
    free(producer_ids);
  }
  delete r;
  for (unsigned tbs = 0; tbs < max_prod_ind; tbs++) {
    sql[0] = '\0';
    for (unsigned it = 0; it < max_item_ind; it++) {
      if (to_be_sent[tbs][it] > 0)
	assert(sprintf(&sql[strlen(sql)], "%d %d ", it + 1, to_be_sent[tbs][it]) < BUF_SIZE - strlen(sql));
    }
    if (sql[0] != '\0') {
      redisReply *reply = (redisReply *)redis_cmd(conn, "LPUSH %s%d %s", QUEUE_PI_BASE, tbs + 1, sql);
      freeReplyObject(reply);
    }
  }
  for (unsigned tbs = 0; tbs < max_prod_ind; tbs++)
    free(to_be_sent[tbs]);
  free(to_be_sent);
}

pqxx::result *get_result(redisContext *conn, pqxx::connection *C, int type, char *sql)
{
  if (type == 3)
    assert(sprintf(sql, "SELECT item_id, o.id, o.how_many, it.how_many FROM order_from_customer o INNER JOIN item it ON item_id = it.id WHERE date_processed IS NULL OR date_ready IS NULL ORDER BY item_id, o.id") < BUF_SIZE);
  else if (type == 4)
    assert(sprintf(sql, "SELECT id, how_many_min - how_many FROM item WHERE how_many < how_many_min") < BUF_SIZE);
  else
    assert(sprintf(sql, "SELECT item_id, id, how_many FROM order_from_customer WHERE date_ready IS NOT NULL AND date_shipped is NULL") < BUF_SIZE);
  pqxx::nontransaction N(*C);
  pqxx::result *r = db_cmd(conn, (pqxx::work *)0, &N, 0 + 4*DB_ONLINE_STATS, sql);
  return r;
}

redisReply **fill_further_ids(redisContext *conn, pqxx::connection *C, int type, unsigned *further_ids_len)
{
  static char sql[1 << 21];
  redisReply **further_ids = (redisReply **)0;
  {
    assert(sprintf(sql, "SELECT channel_code FROM %s WHERE terminated IS NOT NULL AND may_have_nonempty_queue", type == 1? "producer" : "shipper") < BUF_SIZE);
    pqxx::nontransaction N(*C);
    pqxx::result *r = db_cmd(conn, (pqxx::work *)0, &N, 0 + 4*DB_ONLINE_STATS, sql);
    *further_ids_len = 0;
    if (r->size() == 0) {
      delete r;
      return further_ids;
    }
    further_ids = (redisReply **)calloc(r->size(), sizeof(redisReply *));
    for (pqxx::result::const_iterator c = r->begin(); c != r->end(); ++c, (*further_ids_len)++) {
      further_ids[*further_ids_len] = (redisReply *)redis_cmd(conn, "LRANGE %s%d 0 -1", type == 1? QUEUE_PI_BASE : QUEUE_SI_BASE, c[0].as<int>());
    }
    delete r;
  }
  assert(sprintf(sql, "UPDATE %s SET may_have_nonempty_queue = false WHERE terminated IS NOT NULL AND may_have_nonempty_queue", type == 1? "producer" : "shipper") < BUF_SIZE);
  pqxx::work W(*C);
  db_cmd(conn, &W, (pqxx::nontransaction *)0, 2 + 4*DB_ONLINE_STATS, sql);
  return further_ids;
}
